__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from .events import SushiCommandResponse
import grpc
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from . import sushierrors
from . import grpc_gen
from . import sushi_info_types as info_types
from typing import List

####################################
# Sushi program controller class #
####################################

class ProgramController:
    """
    A class to control the program in sushi via gRPC. It controls the programs for the processors and
    can set programs and the get the current program as well as other programs availble.

    Attributes:
        _stub (ProgramControllerStub): Connection stubs to the gRPC program interface implemented in sushi.
    """
    def __init__(self,
                 parent: 'SushiController',
                 address = 'localhost:51051',
                 sushi_proto_def = '/usr/share/sushi/sushi_rpc.proto'):
        """
        The constructor for the ProgramController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError("Parameter address = {}. Should be a string containing the ip-address and port of sushi ('ip-address:port')".format(address)) from e

        self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)
        self._stub = self._sushi_grpc.ProgramControllerStub(channel)
        self._parent = parent

    def get_processor_current_program(self, processor_identifier: int) -> int:
        """
        Get the id of the current program of the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the current program id from.

        Returns:
            int: The id of the processors current program.
        """
        try:
            cr = self._stub.GetProcessorCurrentProgram(self._sushi_proto.ProcessorIdentifier(
                id = processor_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get current program for processor {processor_identifier}")
            return response.program

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}".format(processor_identifier))

    def get_processor_current_program_name(self, processor_identifier: int) -> str:
        """
        Get the name of the current program of the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the current program name from.

        Returns:
            str: The name of the processors current program.
        """
        try:
            cr = self._stub.GetProcessorCurrentProgramName(self._sushi_proto.ProcessorIdentifier(
                id = processor_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get current program name for processor {processor_identifier}")
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}".format(processor_identifier))

    def get_processor_program_name(self, processor_identifier: int, program_identifier: int) -> str:
        """
        Get the name of the specified program on the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the program name from.
            program_identifier (int): The id of the program to get the name of.

        Returns:
            str: The name of the program matching the processor and program id.
        """
        try:
            cr = self._stub.GetProcessorProgramName(self._sushi_proto.ProcessorProgramIdentifier(
                processor = self._sushi_proto.ProcessorIdentifier(id = processor_identifier),
                program = program_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get program {program_identifier} name for processor {processor_identifier}")
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, program id: {}".format(processor_identifier, program_identifier))

    def get_processor_programs(self, processor_identifier: int) -> List[info_types.ProgramInfo]:
        """
        Get a list of the available programs of the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get the programs from.

        Returns:
            List[info_types.ProgramInfo]: A list of the programs available to the processor matching the id.
        """
        try:
            cr = self._stub.GetProcessorPrograms(self._sushi_proto.ProcessorIdentifier(
                id = processor_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Get programs for processor {processor_identifier}")

            program_info_list = []
            for program_info in response.programs:
                program_info_list.append(info_types.ProgramInfo(program_info))

            return program_info_list

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}".format(processor_identifier))

    def set_processor_program(self, processor_identifier: int, program_identifier: int) -> SushiCommandResponse:
        """
        Set the program of the specified processor to the one matching the specified program id.

        Parameters:
            processor_identifier (int): The id of the processor to set the program of.
            program_identifier (int): The id of the program to set.

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            cr = self._stub.SetProcessorProgram(self._sushi_proto.ProcessorProgramSetRequest(
                processor = self._sushi_proto.ProcessorIdentifier(id = processor_identifier),
                program = self._sushi_proto.ProgramIdentifier(program = program_identifier)
            ))
            return self._parent.validate_command_response(cr, f"Set processor {processor_identifier} program {program_identifier}")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}, program id: {}".format(processor_identifier, program_identifier))



__author__ = "Gustav Andersson "
__copyright__ = """

    Copyright 2017-2022 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

import grpc

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from .events import SushiCommandResponse
from . import sushierrors
from . import grpc_gen


##################################
# Sushi session controller class #
##################################


class SessionController:
    """
    A class to save and restore the full state of a sushi session.

    Attributes:
        _stub (SessionControllerStub): Connection stubs to the session interface implemented in sushi.
    """

    def __init__(
        self,
        parent: 'SushiController',
        address: str='localhost:51051',
        sushi_proto_def: str='/usr/share/sushi/sushi_rpc.proto',
        sushi_proto = None, sushi_grpc = None) -> None:
        """
        The constructor for the SessionController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError(
                "Parameter address = {}. Should be a string containing the ip-address and port of sushi ('ip-address:port')".format(
                    address
                )
            ) from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)
        self._stub = self._sushi_grpc.SessionControllerStub(channel)
        self._parent = parent

    def save_binary_session(self) -> bytes | None:
        """
        Save the sushi session.

        Returns:
            bytes: A bytes object containing the complete state of the sushi session.
        """
        try:
            response = self._stub.SaveSession(
                self._sushi_proto.GenericVoidValue())
            return response.SerializeToString()

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def restore_binary_session(self, binary_session: bytes) -> SushiCommandResponse:
        """
        Restore the sushi session from a previously save session state. This will clear all track and loaded plugins

        Returns:
            SushiCommandResponse: an asyncio.Event that will be set asynchronously by elkpy when
            the command has been completed.
            The user MAY await it in an asynchronous program, or check .is_set() on it in a
            loop.
        """
        try:
            grpc_state = self._sushi_proto.SessionState()
            grpc_state.ParseFromString(binary_session)
            cr = self._stub.RestoreSession(grpc_state)
            return self._parent.validate_command_response(cr, "Restore session")
        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

import grpc

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from elkpy.sushicontroller import SushiController

from . import sushierrors
from . import grpc_gen


####################################
# Sushi timing controller class #
####################################

class TimingController:
    """
    A class to control the timing in sushi via gRPC. It can get and reset the different timing statistics
    provided by sushi.

    Attributes:
        _stub (TimingControllerStub): Connection stubs to the gRPC timing interface implemented in sushi.
    """
    def __init__(self,
                 parent,
                 address = 'localhost:51051',
                 sushi_proto_def = '/usr/share/sushi/sushi_rpc.proto',
                 sushi_proto = None, sushi_grpc = None):
        """
        The constructor for the TimingController class setting up the gRPC connection with sushi.

        Parameters:
            address (str): 'ip-addres:port' The ip-addres and port at which to connect to sushi.
            sushi_proto_def (str): path to .proto file with SUSHI's gRPC services definition
        """
        try:
            channel = grpc.insecure_channel(address)
        except AttributeError as e:
            raise TypeError("Parameter address = {}. Should be a string containing the ip-address and port of sushi ('ip-address:port')".format(address)) from e

        if sushi_proto and sushi_grpc:
            self._sushi_proto = sushi_proto
            self._sushi_grpc = sushi_grpc
        else:
            self._sushi_proto, self._sushi_grpc = grpc_gen.modules_from_proto(sushi_proto_def)

        self._stub = self._sushi_grpc.TimingControllerStub(channel)
        self._parent: "SushiController" = parent

    def get_timings_enabled(self) -> bool | None:
        """
        Get the state of timing statstics.

        Returns:
            bool: True if statistics is enabled, False if not.
        """
        try:
            response = self._stub.GetTimingsEnabled(self._sushi_proto.GenericVoidValue())
            return response.value

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def set_timings_enabled(self, enabled: bool) -> None:
        """
        Set the state of timing statstics.

        Parameters:
            bool: True if statistics is enabled, False if not.
        """
        try:
            cr = self._stub.SetTimingsEnabled(self._sushi_proto.GenericBoolValue(value = enabled))
            return self._parent.validate_command_response(cr, context_info=f"Setting timings_enabled to {enabled}")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_engine_timings(self) -> tuple | None:
        """
        Get the average, min and max timings of the engine.
        When Sushi multiprocessor mode is active, gets timings for each thread.

        Timings:
            float: The average engine processing time in ms.
            float: The minimum engine processing time in ms.
            float: The maximum engine processing time in ms.

        Returns:
            a tuple of (main: Timing, threads: list of Timings)
        """
        try:
            response = self._stub.GetEngineTimings(self._sushi_proto.GenericVoidValue())
            return response.main, response.threads

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def get_track_timings(self, track_identifier: int) -> tuple[float, float, float] | None:
        """
        Get the average, min and max timings of the specified track.

        Parameters:
            track_identifier (int): The id of the track to get timings from.

        Returns:
            float: The average track processing time in ms.
            float: The minimum track processing time in ms.
            float: The maximum track processing time in ms.
        """
        try:
            cr = self._stub.GetTrackTimings(self._sushi_proto.TrackIdentifier(
                id = track_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Getting timings for track {track_identifier}")
            return response.timings.average, response.timings.max, response.timings.min

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track id: {}".format(track_identifier))

    def get_processor_timings(self, processor_identifier: int) -> tuple[float, float, float] | None:
        """
        Get the average, min and max timings of the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to get timings from.

        Returns:
            float: The average processor processing time in ms.
            float: The minimum processor processing time in ms.
            float: The maximum processor processing time in ms.
        """
        try:
            cr = self._stub.GetProcessorTimings(self._sushi_proto.ProcessorIdentifier(
                id = processor_identifier
            ))
            response = self._parent.validate_command_response(cr, f"Getting timings for processor {processor_identifier}")
            return response.timings.average, response.timings.max, response.timings.min

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}".format(processor_identifier))

    def reset_all_timings(self):
        """
        Reset all the timings.
        """
        try:
            cr = self._stub.ResetAllTimings(self._sushi_proto.GenericVoidValue())
            return self._parent.validate_command_response(cr, "Resetting all timings.")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e)

    def reset_track_timings(self, track_identifier: int):
        """
        Reset the timings of the specified track.

        Parameters:
            track_identifier (int): The id of the track to reset the timings of.
        """
        try:
            cr = self._stub.ResetTrackTimings(self._sushi_proto.TrackIdentifier(
                id = track_identifier
            ))
            return self._parent.validate_command_response(cr, f"Resetting timings for track {track_identifier}")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With track id: {}".format(track_identifier))

    def reset_processor_timings(self, processor_identifier: int):
        """
        Reset the timings of the specified processor.

        Parameters:
            processor_identifier (int): The id of the processor to reset the timings of.
        """
        try:
            cr = self._stub.ResetProcessorTimings(self._sushi_proto.ProcessorIdentifier(
                id = processor_identifier
            ))
            return self._parent.validate_command_response(cr, f"Resetting timnigs for processor {processor_identifier}")

        except grpc.RpcError as e:
            sushierrors.grpc_error_handling(e, "With processor id: {}".format(processor_identifier))



__author__ = "Ruben Svensson"
__copyright__ = """

    Copyright 2017-2019 Modern Ancient Instruments Networked AB, dba Elk

    elkpy is free software: you can redistribute it and/or modify it under the terms of the
    GNU General Public License as published by the Free Software Foundation, either version 3
    of the License, or (at your option) any later version.

    elkpy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
    even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along with elkpy.  If
    not, see <http://www.gnu.org/licenses/>.
"""
__license__ = "GPL-3.0"

from enum import IntEnum
from types import ModuleType

################
# Custom Enums #
################


class CommandResponseStatus(IntEnum):
    """
    Enum class to hold the values matching the different status code in Command Responses.

    Attributes:
        SUCCESS
        ASYNC_RESPONSE
        ERROR
        UNSUPPORTED_OPERATION
        NOT_FOUND
        OUT_OF_RANGE
        INVALID_ARGUMENTS
    """

    SUCCESS = 1
    ASYNC_RESPONSE = 2
    ERROR = 3
    UNSUPPORTED_OPERATION = 4
    NOT_FOUND = 5
    OUT_OF_RANGE = 6
    INVALID_ARGUMENTS = 7


class PlayingMode(IntEnum):
    """
    Enum class to hold the values matching the different playing modes.

    Attributes:
        STOPPED,
        PLAYING,
        RECORDING
    """

    STOPPED = 1
    PLAYING = 2
    RECORDING = 3


class SyncMode(IntEnum):
    """
    Enum class to hold the values matching the different sync modes.

    Attributes:
        INTERNAL,
        MIDI,
        LINK
    """

    INTERNAL = 1
    MIDI = 2
    LINK = 3


class ParameterType(IntEnum):
    """
    Enum class to hold the values matching the different parameter types.

    Attributes:
        BOOL,
        INT,
        FLOAT,
        STRING_PROPERTY,
        DATA_PROPERTY
    """

    BOOL = 1
    INT = 2
    FLOAT = 3
    STRING_PROPERTY = 4
    DATA_PROPERTY = 5


class PluginType(IntEnum):
    """
    Enum class to hold the values matching the different plugin types.

    Attributes:
        INTERNAL,
        VST2X,
        VST3X,
        LV2
    """

    INTERNAL = 1
    VST2X = 2
    VST3X = 3
    LV2 = 4


class TrackType(IntEnum):
    """
    Enum class to hold the type of track

    Attributes:
        REGULAR,
        PRE,
        POST
    """

    REGULAR = 1
    PRE = 2
    POST = 3


################
# Info Classes #
################


class SushiBuildInfo:
    """
    Class to represent build info from sushi in a clear way

    Attributes:
        version (str): The sushi version
        build_options (List[str]) : A list of build options
        audio_buffer_size (int) : The buffer size used when building sushi
        commit_hash (str) : Commit hash of the build
        build_date (str) : The date sushi was built
    """

    def __init__(self, grpc_SushiBuildInfo: ModuleType) -> None:
        try:
            self.version = grpc_SushiBuildInfo.version
        except AttributeError:
            self._version = ""

        try:
            self.build_options = grpc_SushiBuildInfo.build_options
        except AttributeError:
            self.build_options = [""]

        try:
            self.audio_buffer_size = grpc_SushiBuildInfo.audio_buffer_size
        except AttributeError:
            self.audio_buffer_size = -1

        try:
            self.commit_hash = grpc_SushiBuildInfo.commit_hash
        except AttributeError:
            self.commit_hash = ""

        try:
            self.build_date = grpc_SushiBuildInfo.build_date
        except AttributeError:
            self.build_date = "1-1-1970"

    def __str__(self) -> str:
        s = "{\n"
        s += " version: %s \n" % self.version
        s += " build options: %s \n" % self.build_options
        s += " audio buffer size: %s \n" % self.audio_buffer_size
        s += " commit hash: %s \n" % self.commit_hash
        s += " build date: %s \n" % self.build_date
        s += "}"
        return s

    def __eq__(self, other):
        return (
            self.version == other.version
            and self.build_options == other.build_options
            and self.audio_buffer_size == other.audio_buffer_size
            and self.commit_hash == other.commit_hash
            and self.build_date == other.build_date
        )


class ParameterInfo:
    """
    Class to represent the parameter info received from sushi in a clear way.

    Attributes:
        id (int): The id of the parameter.
        type (str): The type of the parameter.
        label (str): The label of the parameter.
        name (str): The name of the parameter.
        unit (str): The unit of the parameter.
        automatable (bool): If the parameter is automatable or not.
        min_domain_value (float): The minimum value of the parameter.
        max_domain_value (float): The maximum value of the parameter.
    """

    def __init__(self, grpc_ParameterInfo):
        """
        The constructor of the ParameterInfo class.

        Parameters:
            grpc_ParameterInfo (sushi_rpc_pb2.ParameterInfo): the gRPC parameter info object to get the data from.
        """
        try:
            self.id = grpc_ParameterInfo.id
        except AttributeError:
            self.id = 0

        try:
            self.type = ParameterType(grpc_ParameterInfo.type.type)
        except AttributeError:
            self.type = "DUMMY"

        try:
            self.label = grpc_ParameterInfo.label
        except AttributeError:
            self.label = ""

        try:
            self.name = grpc_ParameterInfo.name
        except AttributeError:
            self.name = ""

        try:
            self.unit = grpc_ParameterInfo.unit
        except AttributeError:
            self.unit = ""

        try:
            self.automatable = grpc_ParameterInfo.automatable
        except AttributeError:
            self.automatable = False

        try:
            self.min_domain_value = grpc_ParameterInfo.min_domain_value
        except AttributeError:
            self.min_domain_value = 0.0

        try:
            self.max_domain_value = grpc_ParameterInfo.max_domain_value
        except AttributeError:
            self.max_domain_value = 0.0

    def __str__(self):
        s = "{\n"
        s += " id: %s \n" % self.id
        s += " type: %s \n" % self.type
        s += " label: %s \n" % self.label
        s += " name: %s \n" % self.name
        s += " unit: %s \n" % self.unit
        s += " automatable: %s \n" % self.automatable
        s += " min_domain_value: %s \n" % self.min_domain_value
        s += " max_domain_value: %s \n" % self.max_domain_value
        s += "}"
        return s

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        return (
            self.id == other.id
            and self.type == other.type
            and self.label == other.label
            and self.name == other.name
            and self.unit == other.unit
            and self.automatable == other.automatable
            and self.min_domain_value == other.min_domain_value
            and self.max_domain_value == other.max_domain_value
        )


class PropertyInfo:
    """
    Class to represent the property info received from sushi in a clear way.

    Attributes:
        id (int): The id of the property.
        label (str): The label of the property.
        name (str): The name of the property.
    """

    def __init__(self, grpc_PropertyInfo):
        """
        The constructor of the PropertyInfo class.

        Parameters:
            grpc_PropertyInfo (sushi_rpc_pb2.PropertyInfo): the gRPC parameter info object to get the data from.
        """
        try:
            self.id = grpc_PropertyInfo.id
        except AttributeError:
            self.id = 0

        try:
            self.label = grpc_PropertyInfo.label
        except AttributeError:
            self.label = ""

        try:
            self.name = grpc_PropertyInfo.name
        except AttributeError:
            self.name = ""

    def __str__(self):
        s = "{\n"
        s += " id: %s \n" % self.id
        s += " label: %s \n" % self.label
        s += " name: %s \n" % self.name
        s += "}"
        return s

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other):
        return (
            self.id == other.id
            and self.label == other.label
            and self.name == other.name
        )


class ProcessorInfo:
    """
    Class to represent the processor info received from sushi in a clear way.

    Attributes:
        id (int): The id of the processor.
        label (str): The label of the processor.
        name (str): The name of the processor.
        parameter_count (int): The number of parameters available to the processor.
        program_count (int): The number of programs available to the processor.
    """

    def __init__(self, grpc_ProcessorInfo):
        try:
            self.id = grpc_ProcessorInfo.id
        except AttributeError:
            self.id = 0

        try:
            self.label = grpc_ProcessorInfo.label
        except AttributeError:
            self.label = ""

        try:
            self.name = grpc_ProcessorInfo.name
        except AttributeError:
            self.name = ""

        try:
            self.parameter_count = grpc_ProcessorInfo.parameter_count
        except AttributeError:
            self.parameter_count = 0

        try:
            self.program_count = grpc_ProcessorInfo.program_count
        except AttributeError:
            self.program_count = 0

    def __str__(self):
        s = "{\n"
        s += " id: %s \n" % self.id
        s += " label: %s \n" % self.label
        s += " name: %s \n" % self.name
        s += " parameter count: %s \n" % self.parameter_count
        s += " program count: %s \n" % self.program_count
        s += "}"
        return s

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.id == other.id
            and self.label == other.label
            and self.name == other.name
            and self.parameter_count == other.parameter_count
            and self.program_count == other.program_count
        )


class TrackInfo:
    """
    Class to represent the track info received from sushi in a clear way.

    Attributes:
        id (int): The id of the track.
        label (str): The label of the track.
        name (str): The name of the track.
        channels (int): The number of input channels available to the track.
        buses (int): The number input buses available to the track.
        type (TrackType): The type of track
    """

    def __init__(self, grpc_TrackInfo):
        try:
            self.id = grpc_TrackInfo.id
        except AttributeError:
            self.id = 0

        try:
            self.label = grpc_TrackInfo.label
        except AttributeError:
            self.label = ""

        try:
            self.name = grpc_TrackInfo.name
        except AttributeError:
            self.name = ""

        try:
            self.channels = grpc_TrackInfo.channels
        except AttributeError:
            self.channels = 0

        try:
            self.buses = grpc_TrackInfo.buses
        except AttributeError:
            self.buses = 0

        try:
            self.type = TrackType(grpc_TrackInfo.type.type)
        except Exception:
            self.type = TrackType.REGULAR

        try:
            self.processors = []
            for processor in grpc_TrackInfo.processors:
                self.processors.append(processor.id)
        except AttributeError:
            self.processors = []

    def __str__(self):
        s = "{\n"
        s += " id: %s \n" % self.id
        s += " label: %s \n" % self.label
        s += " name: %s \n" % self.name
        s += " channels: %s \n" % self.channels
        s += " buses: %s \n" % self.buses
        s += " type: %s \n" % self.type
        s += " processors: %s \n" % self.processors
        s += "}"
        return s

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.id == other.id
            and self.label == other.label
            and self.name == other.name
            and self.channels == other.channels
            and self.buses == other.buses
            and self.type == other.type
            and self.processors == other.processors
        )


class ProgramInfo:
    """
    Class to represent the program info received from sushi in a clear way.

    Attributes:
        id (int): The id of the program.
        name (str): The name of the program.
    """

    def __init__(self, grpc_ProgramInfo):
        try:
            self.id = grpc_ProgramInfo.id.program
        except AttributeError:
            self.id = 0

        try:
            self.name = grpc_ProgramInfo.name
        except AttributeError:
            self.name = ""

    def __str__(self):
        s = "{ \n"
        s += " id: %s \n" % self.id
        s += " name: %s \n" % self.name
        s += "}"
        return s

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.id == other.id and self.name == other.name


class ProcessorState:
    """
    Class to represent the processor state info received from sushi in a clear way.

    Attributes:
        program_id (int): The id of the current program.
        bypassed (bool): Whether the processor is currently bypassed or not.
        properties ((int, str)): All property values of the processor.
        parameters ((int, float)): All parameter values of the processor.
        binary_data (bytes): Opaque binary data saved by the plugin.
    """

    def __init__(self, grpc_ProcessorState):
        try:
            self.program_id = grpc_ProcessorState.program_id.value
        except AttributeError:
            self.program_id = None

        try:
            self.bypassed = grpc_ProcessorState.bypassed.value
        except AttributeError:
            self.bypassed = None

        try:
            self.properties = []
            for property in grpc_ProcessorState.properties:
                self.properties.append((property.property.id, property.value))
        except AttributeError:
            self.properties = []

        try:
            self.parameters = []
            for parameter in grpc_ProcessorState.parameters:
                self.parameters.append(
                    (parameter.parameter.parameter_id, parameter.value)
                )
        except AttributeError:
            self.parameters = []

        try:
            self.binary_data = grpc_ProcessorState.binary_data
        except AttributeError:
            self.binary_data = bytes()

    def __str__(self):
        s = "{\n"
        s += " program_id: %s \n" % self.program_id
        s += " bypassed: %s \n" % self.bypassed
        s += " properties: %s \n" % self.properties
        s += " parameters: %s \n" % self.parameters
        s += " binary_data: %s \n" % self.binary_data
        s += "}"
        return s

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.program_id == other.program_id
            and self.bypassed == other.bypassed
            and self.properties == other.properties
            and self.parameters == other.parameters
            and self.binary_data == other.binary_data
        )


class AudioConnection:
    """
    Class to represent an audio connection info received from Sushi in a cleaner way.

    Attributes:
        track (TrackIdentifier): an int
        track_channel (int):
        engine_channel (int):
    """

    def __init__(self, grpc_AudioConnection):
        try:
            self.track = grpc_AudioConnection.track.id
        except AttributeError:
            self.track = 0

        try:
            self.track_channel = grpc_AudioConnection.track_channel
        except AttributeError:
            self.track_channel = 0

        try:
            self.engine_channel = grpc_AudioConnection.engine_channel
        except AttributeError:
            self.engine_channel = 0

    def __str__(self):
        return (
            f"{{ \n track: {self.track}\n"
            f" track_channel: {self.track_channel}\n"
            f" engine_channel: {self.engine_channel}\n}}"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.track == other.track
            and self.track_channel == other.track_channel
            and self.engine_channel == other.engine_channel
        )


class MidiKbdConnection:
    """
    Class to represent a MIDI keyboard connection in Sushi in a cleaner way.

    Attributes:
        track (_sushi_proto.TrackIdentifier): a track id
        channel (_sushi_proto.MidiChannel): a MIDI channel
        port (int): a MIDI port number
        raw_midi (bool): is this track a raw MIDI track or not.
    """

    def __init__(self, grpc_MidiKbdConnection):
        try:
            self.track = grpc_MidiKbdConnection.track.id
        except AttributeError:
            self.track = 0

        try:
            self.channel = grpc_MidiKbdConnection.channel.channel
        except AttributeError:
            self.channel = 0

        try:
            self.port = grpc_MidiKbdConnection.port
        except AttributeError:
            self.port = 0

        try:
            self.raw_midi = grpc_MidiKbdConnection.raw_midi
        except AttributeError:
            self.raw_midi = False

    def __str__(self):
        return (
            f"{{\n track: {self.track}\n"
            f" channel: {self.channel}\n"
            f" port: {self.port}\n"
            f" raw_midi: {self.raw_midi}\n}}"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.track == other.track
            and self.channel == other.channel
            and self.port == other.port
            and self.raw_midi == other.raw_midi
        )


class MidiCCConnection:
    """
    Class to represent a MIDI Continious Controller connection in Sushi in a cleaner way.

    Attributes:
        parameter_id (_sushi_proto.ParameterIdentifier):
        channel (MidiChannel)
        port (int)
        cc_number (int)
        min_range (float)
        max_range (float)
        relative_mode (bool)
    """

    def __init__(self, grpc_MidiCCConnection):
        try:
            self.processor_id = grpc_MidiCCConnection.parameter.processor_id
        except AttributeError:
            self.processor_id = 0

        try:
            self.parameter_id = grpc_MidiCCConnection.parameter.parameter_id
        except AttributeError:
            self.parameter_id = 0

        try:
            self.channel = grpc_MidiCCConnection.channel.channel
        except AttributeError:
            self.channel = 0

        try:
            self.port = grpc_MidiCCConnection.port
        except AttributeError:
            self.port = 0

        try:
            self.cc_number = grpc_MidiCCConnection.cc_number
        except AttributeError:
            self.cc_number = 0

        try:
            self.min_range = grpc_MidiCCConnection.min_range
        except AttributeError:
            self.min_range = 0.0

        try:
            self.max_range = grpc_MidiCCConnection.max_range
        except AttributeError:
            self.max_range = 0.0

        try:
            self.relative_mode = grpc_MidiCCConnection.relative_mode
        except AttributeError:
            self.relative_mode = False

    def __str__(self):
        return (
            f"{{\n parameter: {self.parameter_id}\n"
            f" channel: {self.channel}\n"
            f" port: {self.port}\n"
            f" cc_number: {self.cc_number}\n"
            f" min_range: {self.min_range}\n"
            f" max_range: {self.max_range}\n"
            f" relative_mode: {self.relative_mode}\n}}"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.processor_id == other.processor_id
            and self.parameter_id == other.parameter_id
            and self.channel == other.channel
            and self.port == other.port
            and self.cc_number == other.cc_number
            and self.min_range == other.min_range
            and self.max_range == other.max_range
            and self.relative_mode == other.relative_mode
        )


class MidiPCConnection:
    """
    Class to represent a MIDI Program Change connection in Sushi in a cleaner way.

    Attributes:
        processor (_sushi_proto.ProcessorIdentifier): a processor id
        channel (_sushi_proto.MidiChannel): a MIDI channel
        port (int): a MIDI port number
    """

    def __init__(self, grpc_MidiPCConnection):
        try:
            self.processor = grpc_MidiPCConnection.processor.id
        except AttributeError:
            self.processor = 0

        try:
            self.channel = grpc_MidiPCConnection.channel.channel
        except AttributeError:
            self.channel = 0

        try:
            self.port = grpc_MidiPCConnection.port
        except AttributeError:
            self.port = 0

    def __str__(self):
        return (
            f"{{\n processor: {self.processor}\n"
            f" channel: {self.channel}\n"
            f" port: {self.port}\n}}"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.processor == other.processor
            and self.channel == other.channel
            and self.port == other.port
        )


class CvConnection:
    """
    Class to represent a CV connection in Sushi in a cleaner way

    Attributes:
        parameter_id (int): The id of the connected parameter
        cv_port_id (int): The id of the connected CV port
    """

    def __init__(self, grpc_CvConnection):
        try:
            self.processor_id = grpc_CvConnection.parameter.processor_id
        except AttributeError:
            self.processor_id = 0

        try:
            self.parameter_id = grpc_CvConnection.parameter.parameter_id
        except AttributeError:
            self.parameter_id = 0

        try:
            self.cv_port_id = grpc_CvConnection.cv_port_id
        except AttributeError:
            self.cv_port_id = 0

    def __str__(self):
        return (
            f"{{\n parameter: {self.parameter_id}\n cv_port_id: {self.cv_port_id}\n}}"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.processor_id == other.processor_id
            and self.parameter_id == other.parameter_id
            and self.cv_port_id == other.cv_port_id
        )


class GateConnection:
    """
    Class to represent a Gate connection in Sushi in a cleaner way

    Attributes:
        processor_id (int): The id of the connected processor
        gate_port_id (int): The id of the connected Gate port
        channel (int): The connected midi channel number
        note_no (int): The midi note number to trigger
    """

    def __init__(self, grpc_CvConnection):
        try:
            self.processor_id = grpc_CvConnection.processor.id
        except AttributeError:
            self.processor_id = 0

        try:
            self.gate_port_id = grpc_CvConnection.gate_port_id
        except AttributeError:
            self.gate_port_id = 0

        try:
            self.channel = grpc_CvConnection.channel
        except AttributeError:
            self.channel = 0

        try:
            self.note_no = grpc_CvConnection.note_no
        except AttributeError:
            self.note_no = 0

    def __str__(self):
        return (
            f"{{\n processor: {self.processor_id}\n"
            f" gate_port_id: {self.gate_port_id}\n"
            f" channel: {self.channel}\n"
            f" note_no: {self.note_no}\n}}"
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.processor_id == other.processor_id
            and self.gate_port_id == other.gate_port_id
            and self.channel == other.channel
            and self.note_no == other.note_no
        )
